@@ -10,21 +10,32 @@ use Module::Info;
 use Carp;
 use Error::Simple;
 use File::Spec;
+use Locale::US;
+use CHI;
 
 our %admin1cache;
 our %admin2cache;
 
+#  Some locations aren't found because of inconsistencies in the way things are stored - these are some values I know
+# FIXME: Should be in a configuration file
+my %known_locations = (
+	'Newport Pagnell, Buckinghamshire, England' => {
+		'latitude' => 52.08675,
+		'longitude' => -0.72270
+	},
+);
+
 =head1 NAME
 
 Geo::Coder::Free - Provides a geocoding functionality using free databases of towns
 
 =head1 VERSION
 
-Version 0.04
+Version 0.05
 
 =cut
 
-our $VERSION = '0.04';
+our $VERSION = '0.05';
 
 =head1 SYNOPSIS
 
@@ -66,7 +77,11 @@ sub new {
 
 	my $directory = Module::Info->new_from_loaded(__PACKAGE__)->file();
 	$directory =~ s/\.pm$//;
-	Geo::Coder::Free::DB::init(directory => File::Spec->catfile($directory, 'databases'));
+
+	Geo::Coder::Free::DB::init({
+		directory => File::Spec->catfile($directory, 'databases'),
+		cache => CHI->new(driver => 'Memory', datastore => { })
+	});
 
 	return bless { }, $class;
 }
@@ -81,7 +96,7 @@ sub new {
     # TODO:
     # @locations = $geocoder->geocode('Portland, USA');
     # diag 'There are Portlands in ', join (', ', map { $_->{'state'} } @locations);
- 	
+
 =cut
 
 sub geocode {
@@ -99,6 +114,10 @@ sub geocode {
 	my $location = $param{location}
 		or Carp::croak("Usage: geocode(location => \$location)");
 
+	if($known_locations{$location}) {
+		return $known_locations{$location};
+	}
+
 	my $county;
 	my $state;
 	my $country;
@@ -185,6 +204,13 @@ sub geocode {
 	my @admin2s;
 	my $region;
 	my @regions;
+	# TODO:  Locale::CA for Canadian provinces
+	if(($country =~ /^(United States|USA|US)$/) && (length($county) > 2)) {
+		my $us = Locale::US->new();
+		if(my $twoletterstate = $us->{state2code}{uc($county)}) {
+			$county = $twoletterstate;
+		}
+	}
 	if($county =~ /^[A-Z]{2}/) {
 		# Canadian province or US state
 		$region = $county;
@@ -196,13 +222,13 @@ sub geocode {
 		} elsif(defined($state) && $admin2cache{$state} && !defined($county)) {
 			$region = $admin2cache{$state};
 		} else {
-			@admin2s = @{$self->{'admin2'}->selectall_hashref(asciiname => $county)};
+			@admin2s = $self->{'admin2'}->selectall_hash(asciiname => $county);
 			foreach my $admin2(@admin2s) {
 				if($admin2->{'concatenated_codes'} =~ $concatenated_codes) {
 					$region = $admin2->{'concatenated_codes'};
 					if($region =~ /^[A-Z]{2}\.([A-Z]{2})\./) {
 						my $rc = $1;
-						if($state =~ /^[A-Z]{2}$/) {
+						if(defined($state) && ($state =~ /^[A-Z]{2}$/)) {
 							if($state eq $rc) {
 								$region = $rc;
 								@regions = ();
@@ -222,7 +248,7 @@ sub geocode {
 					$region = $state;
 					@regions = ();
 				} else {
-					@admin2s = @{$self->{'admin2'}->selectall_hashref(asciiname => $state)};
+					@admin2s = $self->{'admin2'}->selectall_hash(asciiname => $state);
 					foreach my $admin2(@admin2s) {
 						if($admin2->{'concatenated_codes'} =~ $concatenated_codes) {
 							$region = $admin2->{'concatenated_codes'};
@@ -236,7 +262,7 @@ sub geocode {
 
 	if((scalar(@regions) == 0) && !defined($region)) {
 		# e.g. Unitary authorities in the UK
-		@admin2s = @{$self->{'admin2'}->selectall_hashref(asciiname => $location)};
+		@admin2s = $self->{'admin2'}->selectall_hash(asciiname => $location);
 		if(scalar(@admin2s) && defined($admin2s[0]->{'concatenated_codes'})) {
 			foreach my $admin2(@admin2s) {
 				if($admin2->{'concatenated_codes'} =~ $concatenated_codes) {
@@ -246,7 +272,7 @@ sub geocode {
 			}
 		} else {
 			# e.g. states in the US
-			my @admin1s = @{$self->{'admin1'}->selectall_hashref(asciiname => $county)};
+			my @admin1s = $self->{'admin1'}->selectall_hash(asciiname => $county);
 			foreach my $admin1(@admin1s) {
 				if($admin1->{'concatenated_codes'} =~ /^$concatenated_codes\./i) {
 					$region = $admin1->{'concatenated_codes'};
@@ -279,7 +305,7 @@ sub geocode {
 
 	# This case nonsense is because DBD::CSV changes the columns to lowercase, wherease DBD::SQLite does not
 	if(wantarray) {
-		my @rc = @{$self->{'cities'}->selectall_hashref($options)};
+		my @rc = $self->{'cities'}->selectall_hash($options);
 		foreach my $city(@rc) {
 			if($city->{'Latitude'}) {
 				$city->{'latitude'} = delete $city->{'Latitude'};
@@ -303,7 +329,7 @@ sub geocode {
 		}
 	}
 
-	if(defined($city) && $city->{'Latitude'}) {
+	if(defined($city) && defined($city->{'Latitude'})) {
 		$city->{'latitude'} = delete $city->{'Latitude'};
 		$city->{'longitude'} = delete $city->{'Longitude'};
 	}
@@ -331,7 +357,7 @@ To be done.
 
 sub reverse_geocode {
 	Carp::croak('Reverse lookup is not yet supported');
-};
+}
 
 =head2	ua
 
@@ -340,7 +366,7 @@ Does nothing, here for compatibility with other geocoders
 =cut
 
 sub ua {
-};
+}
 
 =head1 AUTHOR
 
@@ -359,7 +385,7 @@ VWF, Maxmind and geonames.
 
 =head1 LICENSE AND COPYRIGHT
 
-Copyright 2017 Nigel Horne.
+Copyright 2017-2018 Nigel Horne.
 
 The program code is released under the following licence: GPL for personal use on a single computer.
 All other users (including Commercial, Charity, Educational, Government)
