@@ -1,9 +1,20 @@
 package Geo::Coder::Free::DB;
 
+# Author Nigel Horne: njh@bandsman.co.uk
+# Copyright (C) 2015-2018, Nigel Horne
+
+# Usage is subject to licence terms.
+# The licence terms of this software are as follows:
+# Personal single user, single computer use: GPL2
+# All other users (including Commercial, Charity, Educational, Government)
+#	must apply in writing for a licence for use from Nigel Horne at the
+#	above e-mail.
+
+# Read-only access to databases
+
 use warnings;
 use strict;
 
-use File::Glob;
 use File::Basename;
 use DBI;
 use File::Spec;
@@ -16,6 +27,7 @@ use Error::Simple;
 our @databases;
 our $directory;
 our $logger;
+our $cache;
 
 sub new {
 	my $proto = shift;
@@ -23,9 +35,18 @@ sub new {
 
 	my $class = ref($proto) || $proto;
 
+	if($class eq 'Geo::Coder::Free::DB') {
+		die "$class: abstract class";
+	}
+
+	die "$class: where are the files?" unless($directory || $args{'directory'});
 	# init(\%args);
 
-	return bless { logger => $args{'logger'} || $logger, directory => $args{'directory'} || $directory }, $class;
+	return bless {
+		logger => $args{'logger'} || $logger,
+		directory => $args{'directory'} || $directory,
+		cache => $args{'cache'} || $cache
+	}, $class;
 }
 
 # Can also be run as a class level Geo::Coder::Free::DB::init(directory => '../databases')
@@ -34,6 +55,7 @@ sub init {
 
 	$directory ||= $args{'directory'};
 	$logger ||= $args{'logger'};
+	$cache ||= $args{'cache'};
 	if($args{'databases'}) {
 		@databases = $args{'databases'};
 	}
@@ -74,8 +96,8 @@ sub _open {
 	# Read in the database
 	my $dbh;
 
-	my $directory = $self->{'directory'} || $directory;
-	my $slurp_file = File::Spec->catfile($directory, "$table.sql");
+	my $dir = $self->{'directory'} || $directory;
+	my $slurp_file = File::Spec->catfile($dir, "$table.sql");
 
 	if(-r $slurp_file) {
 		$dbh = DBI->connect("dbi:SQLite:dbname=$slurp_file", undef, undef, {
@@ -86,14 +108,14 @@ sub _open {
 		}
 	} else {
 		my $fin;
-		($fin, $slurp_file) = File::pfopen::pfopen($directory, $table, 'csv.gz:db.gz');
+		($fin, $slurp_file) = File::pfopen::pfopen($dir, $table, 'csv.gz:db.gz');
 		if(defined($slurp_file) && (-r $slurp_file)) {
 			$fin = File::Temp->new(SUFFIX => '.csv', UNLINK => 0);
 			print $fin gunzip_file($slurp_file);
 			$slurp_file = $fin->filename();
 			$self->{'temp'} = $slurp_file;
 		} else {
-			($fin, $slurp_file) = File::pfopen::pfopen($directory, $table, 'csv:db');
+			($fin, $slurp_file) = File::pfopen::pfopen($dir, $table, 'csv:db');
 		}
 		if(defined($slurp_file) && (-r $slurp_file)) {
 			close($fin);
@@ -162,20 +184,16 @@ sub _open {
 			}
 			}
 		} else {
-			$slurp_file = File::Spec->catfile($directory, "$table.xml");
+			$slurp_file = File::Spec->catfile($dir, "$table.xml");
 			if(-r $slurp_file) {
-				# You'll need to install XML::Twig and
-				# AnyData::Format::XML
-				# The DBD::AnyData in CPAN doesn't work - grab a
-				# patched version from https://github.com/nigelhorne/DBD-AnyData.git
-				$dbh = DBI->connect('dbi:AnyData(RaiseError=>1):');
+				$dbh = DBI->connect('dbi:XMLSimple(RaiseError=>1):');
 				$dbh->{'RaiseError'} = 1;
 				if($self->{'logger'}) {
 					$self->{'logger'}->debug("read in $table from XML $slurp_file");
 				}
-				$dbh->func($table, 'XML', $slurp_file, 'ad_import');
+				$dbh->func($table, 'XML', $slurp_file, 'xmlsimple_import');
 			} else {
-				throw Error::Simple("Can't open $directory/$table");
+				throw Error::Simple("Can't open $dir/$table");
 			}
 		}
 	}
@@ -190,6 +208,12 @@ sub _open {
 # Returns a reference to an array of hash references of all the data meeting
 # the given criteria
 sub selectall_hashref {
+	my @rc = selectall_hash(@_);
+	return \@rc;
+}
+
+# Returns an array of hash references
+sub selectall_hash {
 	my $self = shift;
 	my %params = (ref($_[0]) eq 'HASH') ? %{$_[0]} : @_;
 
@@ -200,9 +224,9 @@ sub selectall_hashref {
 
 	if((scalar(keys %params) == 0) && $self->{'data'}) {
 		if($self->{'logger'}) {
-			$self->{'logger'}->trace("$table: selectall_hashref fast track return");
+			$self->{'logger'}->trace("$table: selectall_hash fast track return");
 		}
-		return $self->{'data'};
+		return @{$self->{'data'}};
 	}
 
 	my $query = "SELECT * FROM $table";
@@ -217,16 +241,28 @@ sub selectall_hashref {
 		push @args, $params{$c1};
 	}
 	if($self->{'logger'}) {
-		$self->{'logger'}->debug("selectall_hashref $query: " . join(', ', @args));
+		$self->{'logger'}->debug("selectall_hash $query: " . join(', ', @args));
 	}
 	my $sth = $self->{$table}->prepare($query);
 	$sth->execute(@args) || throw Error::Simple("$query: @args");
+
+	my $key = "$query " . join(', ', @args);
+	my $c;
+	if($c = $self->{cache}) {
+		if(my $rc = $c->get($key)) {
+			return @{$rc};
+		}
+	}
 	my @rc;
-	while (my $href = $sth->fetchrow_hashref()) {
+	while(my $href = $sth->fetchrow_hashref()) {
 		push @rc, $href;
+		last if(!wantarray);
+	}
+	if($c && wantarray) {
+		$c->set($key, \@rc, '1 hour');
 	}
 
-	return \@rc;
+	return @rc;
 }
 
 # Returns a hash reference for one row in a table
@@ -239,7 +275,7 @@ sub fetchrow_hashref {
 
 	$self->_open() if(!$self->{table});
 
-	my $query = "SELECT * FROM $table";
+	my $query = "SELECT DISTINCT * FROM $table";
 	my @args;
 	foreach my $c1(keys(%params)) {
 		if(scalar(@args) == 0) {
@@ -250,6 +286,7 @@ sub fetchrow_hashref {
 		$query .= " $c1 = ?";
 		push @args, $params{$c1};
 	}
+	# $query .= ' ORDER BY entry';
 	if($self->{'logger'}) {
 		$self->{'logger'}->debug("fetchrow_hashref $query: " . join(', ', @args));
 	}
@@ -275,7 +312,7 @@ sub execute {
 	my $sth = $self->{$table}->prepare($query);
 	$sth->execute() || throw Error::Simple($query);
 	my @rc;
-	while (my $href = $sth->fetchrow_hashref()) {
+	while(my $href = $sth->fetchrow_hashref()) {
 		push @rc, $href;
 	}
 
@@ -289,8 +326,12 @@ sub updated {
 	return $self->{'_updated'};
 }
 
-# Return the contents of an arbiratary column in the database which match the given criteria
-# Returns an array of the matches, or just the first entry when called in scalar context
+# Return the contents of an arbiratary column in the database which match the
+#	given criteria
+# Returns an array of the matches, or just the first entry when called in
+#	scalar context
+
+# Set distinct to 1 if you're after a uniq list
 sub AUTOLOAD {
 	our $AUTOLOAD;
 	my $column = $AUTOLOAD;
@@ -308,9 +349,17 @@ sub AUTOLOAD {
 
 	my %params = (ref($_[0]) eq 'HASH') ? %{$_[0]} : @_;
 
-	my $query = "SELECT DISTINCT $column FROM $table";
+	my $query;
+	if(wantarray && !delete($params{'distinct'})) {
+		$query = "SELECT $column FROM $table";
+	} else {
+		$query = "SELECT DISTINCT $column FROM $table";
+	}
 	my @args;
 	foreach my $c1(keys(%params)) {
+		if(!defined($params{$c1})) {
+			$self->{'logger'}->debug("AUTOLOAD params $c1 isn't defined");
+		}
 		# $query .= " AND $c1 LIKE ?";
 		if(scalar(@args) == 0) {
 			$query .= ' WHERE';
@@ -322,7 +371,7 @@ sub AUTOLOAD {
 	}
 	$query .= " ORDER BY $column";
 	if($self->{'logger'}) {
-		if(scalar(@args)) {
+		if(scalar(@args) && $args[0]) {
 			$self->{'logger'}->debug("AUTOLOAD $query: " . join(', ', @args));
 		} else {
 			$self->{'logger'}->debug("AUTOLOAD $query");
@@ -331,7 +380,7 @@ sub AUTOLOAD {
 	my $sth = $self->{$table}->prepare($query) || throw Error::Simple($query);
 	$sth->execute(@args) || throw Error::Simple($query);
 
-	if(wantarray()) {
+	if(wantarray) {
 		return map { $_->[0] } @{$sth->fetchall_arrayref()};
 	}
 	return $sth->fetchrow_array();	# Return the first match only
